@@ -2,7 +2,7 @@ import copy
 import torch
 import deepspeed
 import deepspeed.ops.transformer as transformer_inference
-from .replace_policy import HFBertLayerPolicy, MegatronLayerPolicy, HFGPT2LayerPolicy, HFGPTJLayerPolicy
+from .replace_policy import HFBertLayerPolicy, HFGPT2LayerPolicy, HFGPTJLayerPolicy
 from .replace_policy import replace_policies
 from ..constants import INFERENCE_GENERIC_MODE, INFERENCE_SPECIALIZED_MODE
 from ..runtime.weight_quantizer import WeightQuantization
@@ -53,7 +53,7 @@ class ReplaceWithTensorSlicing:
 
     def qkv_copy(self, dst, src):
         if src is None:
-            return src
+            return torch.nn.Parameter(src)
         src_shape = src.shape
         dst_shape = dst.shape
 
@@ -61,7 +61,7 @@ class ReplaceWithTensorSlicing:
 
         if (len(src_shape) == 2 and len(dst_shape) == 2):
             if src_shape[1] == dst_shape[1]:
-                return src
+                return torch.nn.Parameter(src)
 
             self.merge_assert(src_shape[1], dst_shape[1])
             qkv_size = dst_shape[1] // 3
@@ -75,7 +75,7 @@ class ReplaceWithTensorSlicing:
                 torch.cuda.current_device()).contiguous())
         else:
             if src_shape[0] == dst_shape[0]:
-                return src
+                return torch.nn.Parameter(src)
 
             qkv_size = dst_shape[0] // 3
             qkv_split = [torch.split(src_s, qkv_size, dim=0) for src_s in src_split]
@@ -86,11 +86,11 @@ class ReplaceWithTensorSlicing:
             dst.data.copy_(bias_split[self.gpu_index].to(
                 torch.cuda.current_device()).contiguous())
 
-        return dst
+        return torch.nn.Parameter(dst)
 
     def copy(self, dst, src):
         if src is None:
-            return src
+            return torch.nn.Parameter(src)
 
         src_shape = src.shape
         dst_shape = dst.shape
@@ -98,7 +98,7 @@ class ReplaceWithTensorSlicing:
         if (len(src_shape) == 2 and len(dst_shape) == 2):
 
             if src_shape[0] == dst_shape[0] and src_shape[1] == dst_shape[1]:
-                return src
+                return torch.nn.Parameter(src)
 
             if src_shape[0] != dst_shape[0]:
                 self.merge_assert(src_shape[0], dst_shape[0])
@@ -111,13 +111,13 @@ class ReplaceWithTensorSlicing:
                 torch.cuda.current_device()).contiguous())
         else:
             if src_shape[0] == dst_shape[0]:
-                return src
+                return torch.nn.Parameter(src)
 
             bias_split = torch.split(src.data, dst_shape[-1])
             dst.data.copy_(bias_split[self.gpu_index].to(
                 torch.cuda.current_device()).contiguous())
 
-        return dst
+        return torch.nn.Parameter(dst)
 
 
 def replace_transformer_layer(orig_layer_impl,
@@ -129,6 +129,7 @@ def replace_transformer_layer(orig_layer_impl,
                               hidden_size=-1,
                               num_attention_heads=-1,
                               mp_size=1,
+                              training_mp_size=1,
                               mp_group=None,
                               ep_group=None,
                               expert_mp_group=None,
@@ -199,11 +200,11 @@ def replace_transformer_layer(orig_layer_impl,
                 "This is because the attention computation is partitioned evenly among the parallel GPUs."
         from deepspeed.moe.layer import MoE
         moe = False
-        if isinstance(child.mlp, MoE):
+        if hasattr(child, 'mlp') and isinstance(child.mlp, MoE):
             num_experts = child.mlp.num_experts
             moe = True
 
-        attn_linear_layer, qkvw, qkvb, dense_w, dense_b, scale_attention = policy.attention()
+        attn_linear_layer, qkvw, qkvb, dense_w, dense_b, scale_attention, megatron_v2 = policy.attention()
         if not moe or moe_type == 'standard':
             mlp_linear_layer, _h4h_w, _h4h_b, _4hh_w, _4hh_b = policy.mlp()
         else:
@@ -263,14 +264,19 @@ def replace_transformer_layer(orig_layer_impl,
                     global_experts=num_experts,
                     mlp_type=moe_type)
             else:
+                rotary_dim = config.rotary_dim if hasattr(config, 'rotary_dim') else child.attention.rotary_ndims \
+                                            if hasattr(child, 'attention') and hasattr(child.attention,'rotary_ndims') else -1
                 transformer_config = transformer_inference.DeepSpeedInferenceConfig(
                     hidden_size=hidden_size,
                     heads=num_attention_heads,
                     layer_norm_eps=config.layer_norm_eps if hasattr(
                         config,
-                        'layer_norm_eps') else (config.layer_norm_epsilon if hasattr(
-                            config,
-                            'layer_norm_epsilon') else 1e-12),
+                        'layer_norm_eps') else
+                    (config.layer_norm_epsilon
+                     if hasattr(config,
+                                'layer_norm_epsilon') else config.layernorm_epsilon
+                     if hasattr(config,
+                                'layernorm_epsilon') else 1.0e-12),
                     fp16=fp16,
                     pre_layer_norm=preln,
                     mp_size=mp_size,
@@ -282,9 +288,9 @@ def replace_transformer_layer(orig_layer_impl,
                                                 'attention_layers') else False),
                     window_size=(config.window_size if hasattr(config,
                                                                'window_size') else 1),
-                    rotary_dim=(config.rotary_dim if hasattr(config,
-                                                             'rotary_dim') else -1),
-                    mlp_after_attn=(policy_cls is not HFGPTJLayerPolicy))
+                    rotary_dim=rotary_dim,
+                    mlp_after_attn=(rotary_dim is None or rotary_dim < 0),
+                    training_mp_size=training_mp_size)
 
             if quantize and quantize_settings is not None:
                 (quantization_scales,
@@ -353,6 +359,43 @@ def replace_transformer_layer(orig_layer_impl,
                 qkvw.data = transpose(qkvw.data)
                 dense_w.data = transpose(dense_w.data)
 
+            if megatron_v2:
+                new_module.config.rotate_half = True
+                new_module.config.rotate_every_two = False
+
+                def _transpose(x):
+                    num_attention_heads_per_partition = transformer_config.heads // transformer_config.mp_size
+                    attention_head_size = x.shape[-1] // num_attention_heads_per_partition
+                    new_x_shape = x.size()[:-1] + (num_attention_heads_per_partition,
+                                                   attention_head_size)
+                    x_1 = x.view(*new_x_shape)
+                    (q,
+                     k,
+                     v) = torch.split(x_1,
+                                      (x_1.shape[-1] // 3),
+                                      dim=(x_1.dim() - 1))
+                    if len(q.shape) > 2:
+                        return torch.cat((q.reshape(q.shape[0],
+                                                    -1),
+                                          k.reshape(q.shape[0],
+                                                    -1),
+                                          v.reshape(q.shape[0],
+                                                    -1)),
+                                         dim=-1).reshape(x.shape)
+                    else:
+                        return torch.cat((q.reshape(-1),
+                                          k.reshape(-1),
+                                          v.reshape(-1)),
+                                         dim=-1).reshape(x.shape)
+
+                qkvw = torch.nn.Parameter(_transpose(qkvw).contiguous())
+                qkvb = torch.nn.Parameter(_transpose(qkvb).contiguous())
+
+            dense_b = dense_b * (transformer_config.training_mp_size /
+                                 transformer_config.mp_size)
+            _4hh_b = _4hh_b * (transformer_config.training_mp_size /
+                               transformer_config.mp_size)
+
             if mlp_linear_layer:
                 _h4h_w = [transpose(moe_w1.data)
                           for moe_w1 in _h4h_w] if moe else transpose(_h4h_w.data)
@@ -670,7 +713,10 @@ def replace_module(model, orig_class, replace_fn, _replace_policy):
         for plcy in replace_policies:
             # instantiate a throw-away policy in order to populate the _orig_layer_class
             _ = plcy(None)
-            if plcy._orig_layer_class is not None:
+            if isinstance(plcy._orig_layer_class, list):
+                for orig_layer_class in plcy._orig_layer_class:
+                    policy.update({orig_layer_class: (replace_fn, plcy)})
+            elif plcy._orig_layer_class is not None:
                 policy.update({plcy._orig_layer_class: (replace_fn, plcy)})
     assert len(policy.items()) > 0,\
         "No default policy found! Please specify your policy injection_policy (like {BertLayer:HFBEertLayerPolicy})." +\
@@ -680,6 +726,9 @@ def replace_module(model, orig_class, replace_fn, _replace_policy):
     return replaced_module
 
 
+from ..pipe import PipelineModule
+
+
 def _replace_module(model, policies, layer_id=0):
     """ Traverse model's children recursively and apply any transformations in ``policies``.
     Arguments:
@@ -690,12 +739,14 @@ def _replace_module(model, policies, layer_id=0):
     """
     for name, child in model.named_children():
         if child.__class__ in policies:
-            setattr(
-                model,
-                name,
-                policies[child.__class__][0](child,
-                                             policies[child.__class__][-1],
-                                             layer_id))
+            replaced_module = policies[child.__class__][0](child,
+                                                           policies[child.__class__][-1],
+                                                           layer_id)
+            setattr(model, name, replaced_module)
+            if isinstance(model, PipelineModule):
+                assert hasattr(model, 'forward_funcs'),\
+                    "we require pipe-module to have the list of fwd_functions"
+                model.forward_funcs[model.fwd_map[name]] = replaced_module
             layer_id += 1
         else:
             _, layer_id = _replace_module(child, policies, layer_id=layer_id)
