@@ -18,7 +18,7 @@ Copyright 2021 The Microsoft DeepSpeed Team
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """
- Support different forms of parallelism in DeepSped using multiple process groups.
+ Support different forms of parallelism in DeepSpeed using multiple process groups.
  Given that there are multiple scenarios and use-cases, this file is going to be updated
  frequently. For now, the group creation needed for the training scenario is being implemented.
  For inference and other new scenarios, the code will be either reused or added to this file.
@@ -157,6 +157,52 @@ def _create_expert_and_data_parallel(ep_size):
                 _EXPERT_PARALLEL_GROUP[group_name] = group
 
 
+def _get_expert_parallel_ranks(world_size, model_parallel_size_, expert_parallel_size_):
+    """Generate expert parallel and expert data parallel group ranks list.
+
+        Example - E + M + D parallel
+        world_size = 16
+        model_degree = 2
+        expert_degree = 4 # number of experts in same group
+        mp_group = [0, 1], [2,3], [4,5] ...
+        data_parallel_group =[0,2,4,6,8,10, 12,14],                 [1,3,5,7,9,11,13,15]
+        expert_parallel_group = [0,2,4,6], [8,10,12,14]             [1,3,5,7], [9,11,13,15]
+        expert_data_parallel_group = [0,8],[2,10],[4,12],[6,14],    [1,9],[3,11],[5,13],[7,15]
+
+    Args:
+        world_size (int): Distributed world size.
+        model_parallel_size_ (int): Model parallel group size.
+        expert_parallel_size_ (int): Expert parallel group size.
+
+    Returns:
+        Expert parallel group ranks and Expert data parallel group ranks list.
+    """
+    _ensure_divisibility(world_size, model_parallel_size_)
+    dp_world_size = world_size // model_parallel_size_
+    _ensure_divisibility(dp_world_size, expert_parallel_size_)
+
+    # Generate data parallel groups
+    data_parallel_groups = []
+    dp_group_size = model_parallel_size_
+    for i in range(dp_group_size):
+        data_parallel_groups.append(list(range(i, world_size, dp_group_size)))
+
+    expert_parallel_groups = []
+    expert_data_parallel_groups = []
+    for dp_ranks in data_parallel_groups:
+        # partition of expert parallel groups, e.g. [0,2,4,6], [8,10,12,14]
+        part_ep_groups = []
+        for i in range(0, dp_world_size, expert_parallel_size_):
+            part_ep_groups.append(dp_ranks[i:i + expert_parallel_size_])
+        expert_parallel_groups.extend(part_ep_groups)
+
+        # zip part_ep_groups get expert data parallel ranks, e.g [0,8],[2,10],[4,12],[6,14]
+        for expert_dp_ranks in zip(*part_ep_groups):
+            expert_data_parallel_groups.append(list(expert_dp_ranks))
+
+    return expert_parallel_groups, expert_data_parallel_groups
+
+
 def _create_expert_data_and_model_parallel(expert_parallel_size_, mpu):
     """
         Create expert and data parallel groups based on MPU (model parallel) group.
@@ -170,10 +216,9 @@ def _create_expert_data_and_model_parallel(expert_parallel_size_, mpu):
         mp_group = [0, 1], [2,3], [4,5] ...
         data_parallel_group =[0,2,4,6,8,10, 12,14],                 [1,3,5,7,9,11,13,15]
         expert_parallel_group = [0,2,4,6], [8,10,12,14]             [1,3,5,7], [9,11,13,15]
-        expert_data_parallel_group = [0,8],[2,10],[4,12],[6,14],    [1,9],[3,11],[5,13],[]
+        expert_data_parallel_group = [0,8],[2,10],[4,12],[6,14],    [1,9],[3,11],[5,13],[7,15]
     """
     assert torch.distributed.is_initialized(), "torch distributed is not initialized"
-    assert mpu.model_parallel_is_initialized(), "model parallel group is not initialized"
     model_parallel_size_ = mpu.get_model_parallel_world_size()
 
     world_size = torch.distributed.get_world_size()
@@ -199,22 +244,17 @@ def _create_expert_data_and_model_parallel(expert_parallel_size_, mpu):
     # Only create groups if they don't already exist
     # Need to check conditions outside the group creation loop because of the way torch.dist group creation works
     if group_name not in _EXPERT_DATA_PARALLEL_GROUP and group_name not in _EXPERT_PARALLEL_GROUP:
-        for j in range(model_parallel_size_):
-            for i in range(expert_parallel_size_):
-                ranks = range(i * model_parallel_size_ + j,
-                              world_size,
-                              expert_parallel_size_ * model_parallel_size_)
-                group = torch.distributed.new_group(ranks)
-                if rank in list(ranks):
-                    _EXPERT_DATA_PARALLEL_GROUP[group_name] = group
+        expert_parallel_groups, expert_data_parallel_groups = _get_expert_parallel_ranks(
+            world_size, model_parallel_size_, expert_parallel_size_)
+        for ranks in expert_parallel_groups:
+            group = torch.distributed.new_group(ranks)
+            if rank in list(ranks):
+                _EXPERT_PARALLEL_GROUP[group_name] = group
 
-                for i in range(dp_world_size // expert_parallel_size_):
-                    ranks = range(i * num_ep * model_parallel_size_ + j,
-                                  (i + 1) * expert_parallel_size_ * model_parallel_size_,
-                                  model_parallel_size_)
-                    group = torch.distributed.new_group(ranks)
-                    if rank in list(ranks):
-                        _EXPERT_PARALLEL_GROUP[group_name] = group
+        for ranks in expert_data_parallel_groups:
+            group = torch.distributed.new_group(ranks)
+            if rank in list(ranks):
+                _EXPERT_DATA_PARALLEL_GROUP[group_name] = group
 
 
 def _get_max_expert_size():
